# -*- coding: utf-8 -*-
"""
Created on Thu Jun  1 11:11:57 2017

@author: hector
"""
import datetime
import time
import logging
import json


logger = logging.getLogger(__name__)


class SensorInfo(object):
    def __init__(self):
        self.time_at_sensor = None
        self.time_at_controller = None
        self.timeError = False
        self.status = "UNKNOWN"
        self.name = "Unknown Name"
        self.model = "Unknown Model"
        self.serial = "0"
        self.campaign = None
        self.target_campaign = None
        self.statusInfo = {}
        self.lastError = "None"
        self.softwareInfo = None
        self.sensorType = None
        self.error_counter = 0
        self.error_threshold = 5
        self.uptime = None

    def to_dict(self):
        """ Returns a dictionary representation of the sensor object. """
        dic = {"Status": self.status,
               "Name": self.name,
               "Model": self.model,
               "Serial": self.serial,
               "Task": self.campaign,
               "Target": self.target_campaign,
               "LastError": self.lastError,
               "SoftwareInfo": self.softwareInfo,
               "SensorType": self.sensorType,
               "ErrorCounter": self.error_counter,
               "ErrorThreshold": self.error_threshold,
               "ProcessCPU": self.statusInfo.get("ProcessCPU", "-1"),
               "ProcessID": self.statusInfo.get("ProcessID", "-1"),
               "ProcessMemory": self.statusInfo.get("ProcessMemory", "-1"),
               "StorageSpace": self.statusInfo.get("FreeSpace", "-1"),
               "Temperature": self.statusInfo.get("Temperature", "-273.15"),
               "Timing": self.statusInfo.get("Timing", "UNK"),
               "TimeError": self.timeError}
        return dic

    def to_json(self):
        """ Returns JSON representation of the sensor object. """
        return json.dumps(self.to_dict())

    def check_responsiveness(self, threshold: int):
        """Checks whether the last message from the sensor was too long ago. If so, sets status to "UNKNOWN"
        Requires threshold in seconds."""
        timestamp = int(
                time.mktime(datetime.datetime.now().timetuple()))
        if self.time_at_controller:
            if timestamp - self.time_at_controller > threshold:
                self.set_status_unknown()

    def set_status_unknown(self):
        """Called when engine or conditions render the status unknown."""
        self.status = "UNKNOWN"

    def set_status(self, serial, timestamp, uptime, status, status_info):
        """ Sets attributes as received from the sensor. """
        self.serial = serial
        # All timestamps considered as int seconds from epoch
        self.time_at_sensor = timestamp
        self.uptime = uptime
        self.status = status
        if status == "ERROR":
            self.error_counter += 1
            if status_info and "Cause" in status_info:
                self.lastError = status_info.get("Cause")
            else:
                self.lastError = "Unknown cause"

        if status in ["IDLE", "SENSING"]:
            self.lastError = "None"

        if not status == "BUSY":
            self.statusInfo = status_info
        self._update_controller_time()

    def set_info(self, serial, timestamp, sensor_type, software_info):
        self.serial = serial
        self.time_at_sensor = timestamp
        self.sensorType = sensor_type
        self.softwareInfo = software_info
        self._update_controller_time()

    def _update_controller_time(self):
        """Updates the internal reference for the controller time on this sensor."""
        self.time_at_controller = int(
                time.mktime(datetime.datetime.now().timetuple()))
        if self.status != "OFF" and self._time_offset() > 10:
            logger.warning("Time delta at sensor %s is important: %s.", self.serial, str(self._time_offset()))
            self.timeError = True
        else:
            self.timeError = False

    def set_campaign(self, campaign):
        self.campaign = campaign

    def set_target_campaign(self, campaign):
        self.target_campaign = campaign

    def get_campaign(self):
        return self.campaign

    def _time_offset(self):
        """Returns the time difference between the reported sensor time and the controller."""
        if self.time_at_sensor is None or self.time_at_controller is None:
            return 0
        return self.time_at_controller - self.time_at_sensor

    def is_missing_platform_info(self):
        """ Returns true if this sensor object does not contain the platform info. """
        return self.softwareInfo is None or self.sensorType is None

    def reset_error_counter(self):
        """ Sets error counter to zero. """
        self.error_counter = 0
