"""
ElectroSense Controller
(c) ElectroSense Project

Héctor Cordobés - hector.cordobes@imdea.org - 2017-2019
Markus Fuchs - fuchs@sero-systems.de - 2019-2020
"""

import io
import logging
import sys

import avro.schema
from avro.io import DatumReader, DatumWriter, BinaryEncoder, BinaryDecoder


logger = logging.getLogger(__name__)

messageTypes = ["StatusRequest",
                "SensorStatus",
                "SensorInfo",
                "MeasurementCommand",
                "Log",
                "GroupConfiguration"]


class AvroParser(object):
    # by indicating here the expected names we force both the completeness
    #  of the schema set and the expected names

    def __init__(self, config):
        try:
            with open(config["path"], "r") as f:
                self.avroSchema = avro.schema.Parse(f.read())
        except avro.schema.SchemaParseException as err:
            logger.error("Schema could not be read from %s. Invalid file.", config["path"])
            print(err)
            sys.exit(1)
        except IOError as err:
            logger.error("Schema could not be read from %s. Check AVRO path option.", config["path"])
            print(err)
            sys.exit(1)

    def decode_message(self, msg):
        """Parses an Avro message and returns a dictionary with the parsed message."""
        buf = io.BytesIO(msg)
        decoder = BinaryDecoder(buf)
        reader = DatumReader(self.avroSchema)
        try:
            r = reader.read(decoder)
        except Exception:
            buf.close()
            return {}
        buf.close()  # dirty FIXME TODO
        return r

    def encode_message(self, message_type, data):
        """Create an Avro record with fields type = messageType and message = data
        Encapsulation ensures that the decoder can infer the schema based on the `type` field
        """
        buf = io.BytesIO()
        encoder = BinaryEncoder(buf)
        writer = DatumWriter(self.avroSchema)
        writer.write({"Type": message_type, "Message": data}, encoder)
        msg = bytearray(buf.getvalue())
        buf.close()
        return msg
