# -*- coding: utf-8 -*-
# Sensor v2.0 MQTT flavour
# (c) Electrosense Project
#
#  [TODO] License text
#
# Héctor Cordobés - hcordobes@imdea.org - 2017
# Markus Fuchs - fuchs@sero-systems.de - 2019
#

import datetime
import errno
import json
import logging
import os
import platform
import psutil
import subprocess
from subprocess import check_output
import time

import apt_pkg

from threading import Lock
lock = Lock()

class SensorManager(object):
    """The SensorManager is intended to run on the RaspberryPi sensors and controls the es-sensor process.
    It communicates with the back-end via MQTT."""

    def __init__(self, mqtt_engine, serial_number, cmd_mapper, with_sudo, client_id=None, logger=None):
        self._mqtt = mqtt_engine
        self.serial_number = serial_number
        self.client_id = serial_number if client_id is None else client_id
        self.logger = logger or logging.getLogger(__name__)
        self._cmd_mapper = cmd_mapper
        self._with_sudo = with_sudo

        self.status = "IDLE"
        self.status_info = {}
        self.campaign = "Unknown"
        self.subproc = None
        self._platform_info = rpi_platform_info()

        self._mqtt.register_actions({"MeasurementCommand": self._on_measurement_cmd,
                                     "StatusRequest": self._on_status_request},
                                    {"on": self._on_connect})

        self._mqtt.register_last_will("sensor/status/{}".format(self.client_id),
                                      "SensorStatus",
                                      {"SerialNumber": self.serial_number,
                                       "Timestamp": 0,
                                       "Uptime": 0,
                                       "Status": "OFF"})

    def _on_connect(self):
        self.send_status()
        self.send_info()

    def _on_measurement_cmd(self, msg):
        """Start or Stop measurement
        Args:
            - msg (dict): containing Command, Uuid, Timestamp and SensingParams

        Command may be either START or STOP
          START: sanitycheck, stop previous, create config file, start new campaign
          STOP: sanity check

        When measurement is started, a message indicating this event is published via MQTT
        (status BUSY).
        """
        uuid = msg["Uuid"]
        command = msg["Command"]
        params = msg["SensingParams"]
        self.logger.info("Command: %s", command)
        self.logger.debug("%s", {"Uuid": uuid, "Params": str(params), "Timestamp": str(msg["Timestamp"])})

        # Clean statusInfo
        self.status_info = {"Uuid": uuid}
        if self.status == "ERROR":
            self.logger.warning("Trying to {} after an ERROR status".format(msg["Command"]))

        if msg["Command"] == "START":
            self._on_cmd_start(params)
        elif msg["Command"] == "STOP":
            self._on_cmd_stop()

    def _on_status_request(self, msg):
        if msg["Type"] == "SENSORINFO":
            self.send_info()
            return
        if msg["Type"] == "SENSORSTATUS":
            self.status_info = self._running_info()
            self.send_status()
            return

        self.logger.warning("Invalid StatusRequest Type")
        self.logger.debug(msg)

    def send_status(self):
        self.logger.debug("Generating SensorStatus")
        msg = {"SerialNumber": self.serial_number,
               "Timestamp": current_timestamp(),
               "Uptime": int(time.mktime(datetime.datetime.now().timetuple()) - psutil.boot_time()),
               "Status": self.status,
               "StatusInfo": self.status_info}
        self._mqtt.send_message("SensorStatus", msg, "sensor/status/{}".format(self.client_id))
        self.logger.info("SensorStatus message sent: %s", self.status)

    def send_info(self):
        self.logger.debug("Generating SensorInfo")
        msg = {"SerialNumber": self.serial_number,
               "Timestamp": current_timestamp(),
               "SensorType": "",
               "SoftwareInfo": self._platform_info}
        self._mqtt.send_message("SensorInfo", msg, "sensor/status/{}".format(self.client_id))
        self.logger.info("SensorInfo message sent")

    def send_log(self, level, text):
        self.logger.debug("Generating Log")
        msg = {"SerialNumber": self.serial_number,
               "Timestamp": current_timestamp(),
               "Level": level,
               "LogMessage": text}
        self._mqtt.send_message("Log", msg, "sensor/status/{}".format(self.client_id))
        self.logger.info("Log message sent")

    def start(self):
        """start main loop"""
        while True:
            time.sleep(1)

            if self._mqtt.is_alive() and self.subproc is not None:
                #  is it running?
                output = ""
                error = ""
                output_raw = None
                error_raw = None
                try:
                    output_raw, error_raw = self.subproc.communicate(timeout=1)
                    output = output_raw.decode("utf-8")
                    error = error_raw.decode("utf-8")
                except subprocess.TimeoutExpired:
                    pass
                except ValueError:
                    if output_raw is not None:
                        output = "".join(map(chr, output_raw))
                    if error_raw is not None:
                        error = "".join(map(chr, error_raw))
                except Exception as e:
                    self.logger.error(str(e))

                if len(output) > 0:
                    self.send_log("DEBUG", output)
                    self.logger.debug(output)
                if len(error) > 0:
                    self.send_log("ERROR", error)
                    self.logger.error(error)

                if self.subproc.poll() is not None and not lock.locked():
                    self._on_process_finished(self.subproc.returncode)

    def _error_fail(self, cause, new_status):
        self.status = new_status
        self.status_info["Status"] = "ERROR"
        self.status_info["Cause"] = cause
        self.logger.error(cause)
        # TODO Keeping twice the same info???
        self.campaign = "Unknown"
        self.status_info["CampaignId"] = self.campaign

    def _running_info(self):
        """Get information about running measurement process (PID, Memory, CPU),
        free disk space and campaign id"""
        info = {}

        # Treating the current timesynchronization mechanism as a potentially
        # changing value
        timing_metadata_file = "/tmp/time_sync_mode"

        if os.path.isfile(timing_metadata_file):
            with open(timing_metadata_file) as f:
                lines = [line.rstrip('\n') for line in f]
            if len(lines) == 1:
                info["Timing"] = lines[0]

        info["FreeSpace"] = json.dumps(psutil.disk_usage('/'))
        # temperature = str(psutil.sensors_temperatures())
        if self.subproc is not None:
            if self.subproc.poll() is None:
                p = psutil.Process(pid=self.subproc.pid)
                children = p.children()
                # sudo has the running process as only child
                if len(children) > 0:
                    p = children[0]
                info["ProcessID"] = str(p.pid)
                info["ProcessMemory"] = json.dumps(p.memory_info())
                info["ProcessCPU"] = str(
                    p.cpu_percent(interval=None))

        info["CampaignId"] = self.campaign

        # FIXME platform-specific for RPi
        temp = os.popen("vcgencmd measure_temp").readline().rstrip()
        info["Temperature"] = temp.replace("temp=", "").replace("'C", "")

        return info

    def _on_cmd_start(self, params):
        self.logger.info("Launching sensing process")

        # stop previous campaign if needed
        if self.status == "SENSING":
            self.logger.warning("There was a previous campaign running: %s", str(self.campaign))

        if self.subproc is not None:
            # a process was launched previously
            if not self._stop_process():
                self._error_fail("Could not stop previously running process", "ERROR")
                self.send_status()
                return

        # Send ACK message to controller
        self.status = "BUSY"
        self.send_status()

        command_line = self._cmd_mapper(params)
        # Added feedback info for the executed command
        self.status_info["ExecutedCommand"] = " ".join(command_line)

        try:
            self.subproc = subprocess.Popen(command_line,
                                            cwd="/usr/share/electrosense",
                                            preexec_fn=os.setpgrp,
                                            stdout=subprocess.PIPE,
                                            stderr=subprocess.PIPE)
        except OSError as e:
            self._error_fail("Could not launch process: " + e.strerror, "ERROR")
            self.send_status()
            return

        self.campaign = params.get("CampaignId", "Unknown")
        self.status = "SENSING"
        self.status_info["CampaignId"] = self.campaign
        self.send_status()

    def _on_cmd_stop(self):
        if self.status == "IDLE":
            self.status_info["CampaignId"] = self.campaign
            self.send_status()
            return

        self.status = "BUSY"
        self.send_status()

        if self._stop_process():
            self.status = "IDLE"
            self.campaign = "Unknown"
        else:
            self._error_fail("SEVERE: Could not stop running process.", "ERROR")

        self.send_status()

    def _on_process_finished(self, return_code):
        # This should not happen
        if self.subproc is None:
            self.logger.error("_on_process_finished called and no subproc existed")
            return

        self._stop_process()
        if return_code != 0:
            self._error_fail("Process finished with errors.", "ERROR")
        else:
            self.status = "IDLE"
            self.campaign = "Unknown"
            self.send_status()

    def _stop_process(self):
        """Kill measurement sub process

        Returns:
            True if kill succeeded, False otherwise
        """
        lock.acquire()

        self.logger.debug("Stopping running process...")
        if self.subproc is None:
            self.logger.debug("No previous subprocess found")
            return True

        try:
            # POSIX prevents sending signals to other user process
            # so we cannot kill the parent of a sudo'd process directly from
            # a user-run Python process
            if self._with_sudo:
                exe_name = self.subproc.args[1]
                self.logger.debug("Exec name: %s", exe_name)
                pid = os.getpgid(self.subproc.pid)

                subprocess.call(["sudo", "kill", str(pid)])
                self.logger.debug("Called kill over: %s", str(pid))

                time.sleep(5)
                if self.subproc.poll() is None:
                    self.logger.debug("Tough one, kill -9")
                    subprocess.call(["sudo", "kill", "-9", str(pid)])
            else:
                exe_name = self.subproc.args[0]
                self.subproc.kill()
                time.sleep(1)
                if self.subproc.poll() is None:
                    self.subproc.terminate()

            # Fallback
            if self.subproc.poll() is None:
                self.logger.debug("P-Killing process %s", str(exe_name))
                #subprocess.call(["sudo", "pkill", "-9", exe_name])
                subprocess.call(["sudo", "pkill", exe_name])

        except OSError as e:
            self.logger.debug("OSError killing process: %s", str(e.errno))
            # Non-existent process exception is good
            if e.errno != errno.ESRCH:
                self._error_fail("OS error killing previous sensor instance: {}".format(e.strerror), "ERROR")
                self.send_status()
                return False

        finally:
            lock.release()

        self.subproc = None
        return True


def current_timestamp():
    return int(time.mktime(datetime.datetime.now().timetuple()))


def create_cmd_mapper(with_sudo, parameter_table, logger=None):
    if logger is None:
        logger = logging.getLogger(__name__)

    def cmd_list_from_params(input_parameters):
        """Generate a list of parameters representing the system command to start measurements,
        given a dictionary of parameters as received by the backend.

        Example:
            params = {"command": "run_gpu_sensor", "minFreq": 0, "maxFreq": 500}
            returns ["run_gpu_sensor", "0", "500"]

        The parameters must be present in the [parameters] section of the configuration,
        otherwise they are ignored. Each parameter is substituted with its command line
        argument.
        [2018-09-11] The check for mandatory parameters is left to the process implementation
        and the system call. With this change the logic is simpler for future new processes.
        As the parameter list may not be completely homogeneous a prefix/suffix approach is used.
        """

        # Order is preserved; use lowercase
        prefix_params = ["command"]
        suffix_params = ["minfreq", "maxfreq"]
        input_parameters = dict((k.lower(), v) for k, v in input_parameters.items())

        # Add prefix params if present
        # Add all params which are not prefix or suffix
        # And add suffix params if present
        out_param_keys = [p for p in prefix_params if p in input_parameters.keys()] + \
                         [p for p in input_parameters.keys() if p not in prefix_params and p not in suffix_params] + \
                         [p for p in suffix_params if p in input_parameters.keys()]

        # begin construction of config
        cfg_list = ["sudo"] if with_sudo else []

        # from each parameter name, look for its CLI option
        for param in out_param_keys:
            # FIXME lower() in ConfigParser should be bypassed
            if param in parameter_table.keys():
                try:
                    param_id = parameter_table[param]
                    if param_id:
                        cfg_list.append(param_id)
                    cfg_list.append(input_parameters[param])
                except:
                    logger.error("Boom %s")
            else:
                logger.warning("Unknown option parameter: %s", param)

        logger.debug("Config file contents: %s", str(cfg_list))

        return cfg_list

    return cmd_list_from_params


def rpi_platform_info():
    """Get information about installed electrosense Debian packages, kernel,
    distro, libc, CPU info and image release number.
    Besides ^electrosense.* packages, extra_packages are also considered."""

    rpimodel = {"0002": "Model B Rev 1 256MB",
                "0003": "Model B Rev 1 ECN0001 (no fuses, D14 removed) 256MB",
                "0004": "Model B Rev 2 256MB",
                "0005": "Model B Rev 2 256MB",
                "0006": "Model B Rev 2 256MB",
                "0007": "Model A 256MB",
                "0008": "Model A 256MB",
                "0009": "Model A 256MB",
                "000d": "Model B Rev 2 512MB",
                "000e": "Model B Rev 2 512MB",
                "000f": "Model B Rev 2 512MB",
                "0010": "Model B+ 512MB",
                "0013": "Model B+ 512MB",
                "900032": "Model B+ 512MB",
                "0011": "Compute Module 512MB",
                "0014": "Compute Module 512MB (Embest, China)",
                "0012": "Model A+ 256MB",
                "0015": "Model A+ (Embest, China)",
                "a01041": "Pi 2 Model B v1.1 1GB (Sony, UK)",
                "a21041": "Pi 2 Model B v1.1 1GB (Embest, China)",
                "a22042": "Pi 2 Model B v1.2 1GB",
                "900092": "Pi Zero v1.2 512MB",
                "900093": "Pi Zero v1.3 512MB",
                "9000C1": "Pi Zero W 512MB",
                "a02082": "Pi 3 Model B 1GB (Sony, UK)",
                "a22082": "Pi 3 Model B 1GB (Embest, China)",
                "a020d3": "Pi 3 Model B+ 1GB (Sony, UK)"}
    apt_pkg.init()
    cache = apt_pkg.Cache()
    info = {}
    sensing_package = ["electrosense-sensor"]
    extra_packages = ["websocketctrl"]
    image_metadata_file = "/boot/es_image"
    control_file = "/etc/electrosense/control"
    sensing_version = "Unknown"
    image_version = "Unknown"

    # RaspberryModel
    revision = "0000"
    try:
        with open('/proc/cpuinfo', 'r') as f:
            for line in f:
                if line[0:8] == 'Revision':
                    length = len(line)
                    revision = line[11:length-1]
        info["Model"] = rpimodel[revision]
        info["Name"] = "Raspberry Pi"
    except (OSError, KeyError):
        try:
            with open("/proc/device-tree/model", "r") as f:
                info["Model"] = f.readline()
            info["Name"] = "Raspberry Pi"
        except:
            info["Name"] = "Unknown"
            info["Model"] = "Unknown"
    info["Revision"] = revision

    # Installed packages
    for package in sensing_package:
        if package in cache:
            package_info = cache[package]
            sensing_version = package_info.current_ver.ver_str
            break

    # Obtain all packages named as ^electrosense.* and specific extra packages
    # to extract the package version
    installed_packages = {}
    for package in list(
            filter(lambda x: x.name.startswith("electrosense") or x.name in extra_packages, cache.packages)):
        if package.current_state == 6:
            installed_packages[package.name] = package.current_ver.ver_str

    control_present = os.path.isfile(control_file)

    # If present and valid, obtain image version.
    # If not available, the default value will be used ("Unknown")
    if os.path.isfile(image_metadata_file):
        with open(image_metadata_file) as f:
            lines = [line.rstrip('\n') for line in f if line.startswith("version")]
        if len(lines) > 0:
            image_version = lines[0].replace("version:", "")

    # Obtain whether Converter board is installed or not
    board_output = check_output("lsusb | grep 'Clay Logic' | awk '{print $6}'", shell=True).decode().rstrip()
    if board_output:
        board_model = board_output.split(":")[1]
    else:
        board_model = "null"

    info["ElectrosensePackage"] = sensing_version
    info["InstalledPackages"] = json.dumps(installed_packages)
    info["Official"] = str(control_present)
    info["ImageVersion"] = image_version
    info["Uname"] = json.dumps(platform.uname())
    info["Platform"] = json.dumps(platform.platform())
    ## removed in Python 3.8: info["Distro"] = json.dumps(platform.linux_distribution())
    info["LibC"] = json.dumps(platform.libc_ver())
    info["Cores"] = str(psutil.cpu_count())
    info["ExtensionBoard"] = board_model
    return info
